import asyncio
from typing import Optional

from robotica.executor import Executor
from robotica.plugins import Plugin
from robotica.schedule import Schedule
from robotica.types import Config


class Input(Plugin):
    def __init__(
            self, *,
            name: str,
            loop: asyncio.AbstractEventLoop,
            config: Config,
            executor: Executor,
            schedule: Optional[Schedule]) -> None:
        super().__init__(name=name, loop=loop, config=config)
        self._executor = executor
        self._schedule = schedule



""" Robotica Schedule. """
import asyncio
import logging
from typing import Dict, Set, List  # NOQA

from robotica.plugins.outputs import Output
from robotica.types import Action

logger = logging.getLogger(__name__)


class Executor:
    def __init__(
            self, loop: asyncio.AbstractEventLoop, config: Dict) -> None:
        self._loop = loop
        self._config = config
        self._outputs = []  # type: List[Output]
        self._lock = asyncio.Lock()

    def add_output(self, output: Output) -> None:
        self._outputs.append(output)

    def action_required_for_locations(
            self, locations: Set[str], action: Action) -> Set[str]:

        required_locations = set([
            location
            for output in self._outputs
            for location in locations
            if output.is_action_required_for_location(location, action)
        ])

        return required_locations

    async def do_action(self, locations: Set[str], action: Action) -> None:
        required_locations = self.action_required_for_locations(locations, action)
        if len(required_locations) == 0:
            return

        with await self._lock:
            coros = [
                output.execute(location, action)
                for output in self._outputs
                for location in locations
            ]
            await asyncio.gather(
                *coros,
                loop=self._loop
            )

    async def do_actions(self, locations: Set[str], actions: List[Action]) -> None:
        for action in actions:
            await self.do_action(locations, action)



import asyncio
import logging
from typing import Dict

from robotica.types import Config

logger = logging.getLogger(__name__)


class Plugin:
    def __init__(
            self, *,
            name: str,
            loop: asyncio.AbstractEventLoop,
            config: Config) -> None:
        super().__init__()
        self._name = name
        self._loop = loop
        self._config = config

    def start(self) -> None:
        pass

    def stop(self) -> None:
        pass




import asyncio
import json
import logging
import platform
from typing import Any, Optional, Dict, Tuple, List

from hbmqtt.client import MQTTClient, ClientException, QOS_0

from robotica.executor import Executor
from robotica.plugins.inputs import Input
from robotica.schedule import Schedule
from robotica.types import Config, Action

logger = logging.getLogger(__name__)

JsonType = Any


class MqttInput(Input):
    def __init__(
            self, *,
            name: str,
            loop: asyncio.AbstractEventLoop,
            config: Config,
            executor: Executor,
            schedule: Optional[Schedule]) -> None:
        super().__init__(
            name=name,
            loop=loop,
            config=config,
            executor=executor,
            schedule=schedule,
        )
        self._disabled = self._config['disabled']
        self._broker_url = self._config['broker_url']
        self._locations = self._config.get('locations', []) or []
        self._task = None  # type: Optional[asyncio.Task]
        self._client = MQTTClient()

    def start(self) -> None:
        if not self._disabled:
            self._task = self._loop.create_task(self._mqtt())

    def stop(self) -> None:
        if not self._disabled and self._task is not None:
            self._task.cancel()
            try:
                self._loop.run_until_complete(self._task)
            except asyncio.CancelledError:
                pass

    def _get_topics(self) -> List[Tuple[str, int]]:
        if self._schedule is None:
            topics = [
                ('/action/%s/' % location, QOS_0)
                for location in self._locations
            ]
        else:
            topics = [
                ('/schedule/', QOS_0),
                ('/execute/', QOS_0),
            ]
        return topics

    async def _process_execute(self, data: JsonType) -> None:
        if self._schedule is None:
            return

        reply_topic = data.get('reply_topic', None)
        server = platform.node()

        async def reply(data: JsonType) -> None:
            client = self._client
            raw_data = json.dumps(data).encode('UTF8')
            if reply_topic is not None:
                await client.publish(reply_topic, raw_data, qos=QOS_0)

        try:
            locations = data['locations']
            actions = data['actions']
        except KeyError:
            logger.error("Required value missing.")
            await reply({'status': 'error', 'server': server, })
            return

        try:
            await reply({'status': 'processing', 'server': server, })
            await self._executor.do_actions(locations, actions)
            await reply({'status': 'success', 'server': server, })
        except Exception as e:
            logger.exception("Error in _execute")
            await reply({'status': 'error', 'server': server, })

        return

    async def _process_schedule(self, data: JsonType) -> None:
        if self._schedule is not None:
            await self._schedule.set_schedule(data)
            self._schedule.save_schedule()

    async def _process_action(self, location: str, action: Action) -> None:
        if self._schedule is None:
            await self._executor.do_action({location}, action)

    async def _process(self, topic: str, data: JsonType) -> None:
        logger.info("Received %s %s", topic, data)
        if topic.startswith("/execute/"):
            await self._process_execute(data)
        if topic.startswith("/schedule/"):
            await self._process_schedule(data)
        if topic.startswith("/action/"):
            location = topic[8:].rstrip("/")
            await self._process_action(location, data)

    async def _mqtt(self) -> None:
        client = self._client
        await client.connect(self._broker_url)
        await client.subscribe(self._get_topics())

        while True:
            try:
                message = await client.deliver_message()
                packet = message.publish_packet
                topic = packet.variable_header.topic_name
                raw_data = bytes(packet.payload.data).decode('UTF8')

                try:
                    data = json.loads(raw_data)
                    await self._process(topic, data)
                except json.JSONDecodeError as e:
                    logger.error("JSON Error %s" % e)

            except asyncio.CancelledError:
                topics = self._get_topics()
                await client.unsubscribe([t[0] for t in topics])
                await client.disconnect()
                raise
            except ClientException as e:
                logger.error("Client exception: %s" % e)
            except Exception as e:
                logger.exception("Unknown exception.")



# -*- coding: utf-8 -*-

"""Console script for Robotica."""
import asyncio
import importlib
import logging
from typing import List, Any

import click
import click_log
import yaml

from robotica.executor import Executor
from robotica.plugins import Plugin
from robotica.plugins.inputs import Input
from robotica.plugins.outputs import Output
from robotica.schedule import Schedule

logger = logging.getLogger('robotica')
click_log.basic_config(logger)


def _load_class(class_name: str) -> Any:
    """
    Dynamically load a class from a string
    """

    class_data = class_name.split(".")
    module_path = ".".join(class_data[:-1])
    class_str = class_data[-1]

    class_module = importlib.import_module(module_path)
    return getattr(class_module, class_str)


@click.command()
@click.option('--config', default="config/config.yaml", help='Path to config.')
@click.option('--schedule', default="config/schedule.yaml", help='Path to schedule config or None.')
@click_log.simple_verbosity_option(logger)
def main(config: str, schedule: str) -> None:
    """Console script for robotica."""
    with open(config, "r") as file:
        config_dict = yaml.safe_load(file)
        output_dict = config_dict['outputs']
        input_dict = config_dict['inputs']

    loop = asyncio.get_event_loop()
    plugins = []  # type: List[Plugin]

    executor_obj = Executor(loop, config_dict['executor'])
    for name in output_dict.keys():
        output_plugin_config = output_dict[name]
        output_plugin_class = _load_class(output_plugin_config['plugin'])
        assert issubclass(output_plugin_class, Output)
        output_plugin = output_plugin_class(
            name=name,
            loop=loop,
            config=output_plugin_config,
        )
        output_plugin.start()
        executor_obj.add_output(output_plugin)
        plugins.append(output_plugin)

    if schedule.upper() == "NONE":
        schedule_obj = None
    else:
        schedule_obj = Schedule(schedule, executor_obj)
        schedule_obj.start()

    for name in input_dict.keys():
        input_plugin_config = input_dict[name]

        input_plugin_class = _load_class(input_plugin_config['plugin'])
        assert issubclass(input_plugin_class, Input)
        input_plugin = input_plugin_class(
            name=name,
            loop=loop,
            config=input_plugin_config,
            executor=executor_obj,
            schedule=schedule_obj,
        )
        input_plugin.start()
        plugins.append(input_plugin)

    try:
        loop.run_forever()
    finally:
        for plugin in reversed(plugins):
            plugin.stop()
        pending = asyncio.Task.all_tasks()
        for p in pending:
            p.cancel()
            try:
                loop.run_until_complete(p)
            except asyncio.CancelledError:
                pass
        loop.close()



from robotica.plugins import Plugin
from robotica.types import Action


class Output(Plugin):
    def is_action_required_for_location(self, location: str, action: Action) -> bool:
        raise NotImplemented()

    async def execute(self, location: str, action: Action) -> None:
        raise NotImplemented()



import asyncio
import logging
from typing import Set, Dict

from aiolifxc.aiolifx import Lights, Light, Color, DeviceOffline

from robotica.plugins.outputs import Output
from robotica.types import Action, Config

logger = logging.getLogger(__name__)


class LifxOutput(Output):
    def __init__(
            self, *,
            name: str,
            loop: asyncio.AbstractEventLoop,
            config: Config) -> None:
        super().__init__(
            name=name,
            loop=loop,
            config=config,
        )
        self._disabled = self._config['disabled']
        self._lights = Lights(loop=self._loop)
        self._locations = self._config.get('locations', {}) or {}

    def start(self) -> None:
        if not self._disabled:
            logger.debug("LIFX enabled.")
            self._task = self._lights.start_discover()

    def stop(self) -> None:
        if not self._disabled:
            self._task.cancel()

    def _get_labels_for_location(self, location: str) -> Set[str]:
        labels = set(self._locations.get(location, []))
        return labels

    def is_action_required_for_location(self, location: str, action: Action) -> bool:
        if self._disabled:
            return False

        labels = self._get_labels_for_location(location)
        if len(labels) == 0:
            return False

        if 'lights' in action:
            return True

        return False

    async def execute(self, location: str, action: Action) -> None:
        if 'lights' in action:
            lights = action['lights']

            lights_action = lights['action']
            if lights_action == "flash":
                await self.flash(location=location)
            elif lights_action == "wake_up":
                await self.wake_up(location=location)
            elif lights_action == "turn_off":
                await self.turn_off(location=location)
            else:
                logger.error("Unknown action '%s'.", action)

    def _get_lights_from_location(self, location: str) -> Lights:
        labels = self._get_labels_for_location(location)
        lights = self._lights.get_by_lists(labels=list(labels))  # type: Lights
        return lights

    async def wake_up(self, location: str) -> None:
        async def single_device(device: Light) -> None:
            try:
                power = await device.get_power()
                if not power:
                    await device.set_color(
                        Color(hue=0, saturation=0, brightness=0, kelvin=2500))
                await device.set_power(True)
                await device.set_color(
                    Color(hue=0, saturation=0, brightness=100, kelvin=2500),
                    duration=60000)
            except DeviceOffline:
                logger.error("Light is offline %s.", device)

        lights = self._get_lights_from_location(location)
        logger.info("Lifx wakeup for lights %s.", lights)
        await lights.do_for_every_device(Light, single_device)

    async def flash(self, location: str) -> None:
        lights = self._get_lights_from_location(location)
        logger.info("Lifx flash for lights %s.", lights)
        await lights.set_waveform(
            color=Color(hue=0, saturation=100, brightness=100, kelvin=3500),
            transient=1,
            period=1000,
            cycles=2,
            duty_cycle=0,
            waveform=0,
        )

    async def turn_off(self, location: str) -> None:
        lights = self._get_lights_from_location(location)
        logger.info("Lifx flash for lights %s.", lights)
        await lights.set_light_power(False)



""" Give verbal message. """
import asyncio
import json
import logging
from typing import Dict

from hbmqtt.client import MQTTClient, ClientException, QOS_0

from robotica.plugins.outputs import Output
from robotica.types import JsonType, Action, Config

logger = logging.getLogger(__name__)


class MqttOutput(Output):

    def __init__(
            self, *,
            name: str,
            loop: asyncio.AbstractEventLoop,
            config: Config) -> None:
        super().__init__(
            name=name,
            loop=loop,
            config=config,
        )
        self._disabled = self._config['disabled']
        self._broker_url = self._config['broker_url']
        self._locations = self._config.get('locations', {}) or {}
        self._client = MQTTClient()

    def start(self) -> None:
        if not self._disabled:
            self._loop.run_until_complete(self._client.connect(self._broker_url))

    def stop(self) -> None:
        pass

    def is_action_required_for_location(self, location: str, action: Action) -> bool:
        if self._disabled:
            return False

        if location not in self._locations:
            return False

        return True

    async def execute(self, location: str, action: Action) -> None:
        if not self.is_action_required_for_location(location, action):
            return

        await self._execute(
            '/action/%s/' % location,
            action,
        )

    async def _execute(self, topic: str, data: JsonType) -> None:
        logger.debug("About to publish %r to %s" % (data, topic))
        raw_data = json.dumps(data).encode('UTF8')
        try:
            await self._client.publish(
                topic,
                raw_data,
                qos=QOS_0,
            )
        except ClientException:
            logger.exception("The client operation failed.")
